#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import DREAM
import numpy

def example_01():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 1: make your own probability distribution")
    print("           sample from the Gaussian distribution: f(x) = exp(-x^2)")
    print("           ignoring scaling constants, using 3000 samples")
    print("    See the comments in example_dream_01.cpp\n")

    iNumDimensions = 1
    iNumChains = 30
    iNumBurnupIterations = 200
    iNumCollectIterations = 1000 # total samples are iNumChains x iNumCollectIterations

    state = DREAM.State(iNumChains, GridOrIntDimensions = iNumDimensions)

    # initialize with uniform samples on [-2, 2]
    state.setState(DREAM.genUniformSamples((-2.0,), (2.0,), state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 lambda x : numpy.exp( - x[:,0]**2 ), # Gaussian mean 0.0 variance 0.5
                 DREAM.Domain("unbounded"),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.5),
                 DREAM.DifferentialUpdate(90))

    aMean, aVariance = state.getHistoryMeanVariance()

    print("Using regular form:")
    print("mean    {0:13.6f}   error {1:14.6e}".format(aMean[0], numpy.abs(aMean[0])))
    print("variance{0:13.6f}   error {1:14.6e}".format(aVariance[0], numpy.abs(aVariance[0] -0.5)))

    # reset the state
    state = DREAM.State(iNumChains, GridOrIntDimensions = iNumDimensions)

    # initialize with uniform samples on [-2, 2]
    state.setState(DREAM.genUniformSamples((-2.0,), (2.0,), state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 lambda x : - x[:,0]**2, # Gaussian mean 0.0 variance 0.5
                 DREAM.Domain("unbounded"),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.5),
                 DREAM.DifferentialUpdate(90),
                 typeForm = DREAM.typeLogform)

    aMean, aVariance = state.getHistoryMeanVariance()

    print("Using logarithm form:")
    print("mean    {0:13.6f}   error {1:14.6e}".format(aMean[0], numpy.abs(aMean[0])))
    print("variance{0:13.6f}   error {1:14.6e}".format(aVariance[0], numpy.abs(aVariance[0] -0.5)))


if __name__ == "__main__":
    example_01()



#!@Tasmanian_string_python_hashbang@

# necessary import for every use of TASMANIAN
#
import TasmanianSG
import numpy as np
import math

#from random import uniform

#import matplotlib.pyplot as plt
#import matplotlib.colors as cols
#from mpl_toolkits.mplot3d import Axes3D

###############################################################################
# This file is included for testing, debugging and development reasons
# This file will reside in the cmake build folder, but not the install folder
# Basically, add testing code to this file and it will be automatically
# included as part of the cmake
# You can also use the file without cmake
###############################################################################

print("Add code to this file to test, debug, or develop features")



#!/usr/bin/env python

# The hash-bang is used only but the GNU Make build system
# The CMake build system uses the PYTHON_EXECUTABLE and ignores the hash-bang.

import unittest
import TasmanianSG
import math
import sys, os
import numpy as np
from random import uniform
from ctypes import cdll

import testCommon
import testBasicIO
import testAcceleration
import testExceptions
import testMakeUpdate
import testRefinement
import testUnstructuredData
import testMisc
import testAddons
import testDream

grid = TasmanianSG.TasmanianSparseGrid()

ttc = testCommon.TestTasCommon()

# python-coverage run testTSG.py
# python-coverage html
# python-coverage report

class TestTasmanian(unittest.TestCase):
    def testBasicIO(self):
        print("\nTesting core I/O test")
        tester = testBasicIO.TestTasClass()
        tester.performIOTest()

    def testAcceleratedEvaluate(self):
        print("\nTesting accelerated evaluate consistency")
        tester = testAcceleration.TestTasClass()
        tester.performAccelerationTest()

    def testBasicException(self):
        print("\nTesting error handling")
        tester = testExceptions.TestTasClass()
        tester.performExceptionsTest()

    def testAMakeUpdate(self):
        print("\nTesting core make/update grid")
        tester = testMakeUpdate.TestTasClass()
        tester.performMakeUpdateTests()

    def testBRefinement(self):
        print("\nTesting core refine grid")
        tester = testRefinement.TestTasClass()
        tester.performRefinementTest()

    def testCUnsructuredData(self):
        print("\nTesting core learning from random samples")
        tester = testUnstructuredData.TestTasClass()
        tester.performUnstructuredDataTests()

    def testZMisc(self):
        print("\nTesting plotting and other misc")
        tester = testMisc.TestTasClass()
        tester.performMiscTests()

    def testAddons(self):
        print("\nTesting addon wrappers")
        tester = testAddons.TestTasClass()
        tester.performAddonTests()

    def testDream(self):
        print("\nTesting DREAM wrappers")
        tester = testDream.TestTasClass()
        tester.performDreamTests()

if __name__ == '__main__':
    unittest.main()





#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import DREAM
from Tasmanian import SparseGrid
from Tasmanian import makeSequenceGrid
from Tasmanian import loadNeededPoints
import numpy

def example_04():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 4: similar to Example 3, but the data is noisy and the multi-modal posterior comes")
    print("           from a symmetry in the model")
    print("           set inference problem: identify x_0, x_1, x_2 and x_3 (four) model parameters")
    print("                                  from noisy data")
    print("           model: f(x) = x_0*exp(x_1*t) + x_2*exp(x_3*t),  data: d = exp(t) + 0.4*exp(3*t)")
    print("           note the symmetry between x_1 and x_3 which results in a bi-modal posterior")
    print("           t in [0,1], t is discretized with 64 equidistant nodes")
    print("           likelihood is exp(-64 * (f(x) - d)^2)")
    print("           using a sparse grid to interpolate the model\n")

    iNumDimensions = 4
    iNumOutputs = 64
    iNumChains = 50
    iNumBurnupIterations  = 1000
    iNumCollectIterations = 1000

    def model(aX):
        # using 64 nodes in the interior of (0.0, 1.0)
        t = numpy.linspace(1.0 / 128.0, 1.0 - 1.0 / 128.0, 64)
        return aX[0] * numpy.exp(aX[1] * t) + aX[2] * numpy.exp(aX[3] * t)

    aData = (model([1.0, 1.0, 0.4, 3.0]) # add noise
        + DREAM.genGaussianSamples([0.0 for i in range(iNumOutputs)],
                                   [1.0 / float(iNumOutputs) for i in range(iNumOutputs)],
                                   1).reshape((iNumOutputs,)))

    grid = makeSequenceGrid(iNumDimensions, iNumOutputs, 15, 'iptotal', 'leja')

    domain_lower = numpy.array([0.2, 0.5, 0.2, 0.5])
    domain_upper = numpy.array([1.2, 4.0, 1.2, 4.0])
    grid.setDomainTransform(numpy.column_stack([domain_lower, domain_upper]))

    loadNeededPoints(lambda x, tid : model(x), grid, 1)

    likely = DREAM.LikelihoodGaussIsotropic(1.0 / float(iNumOutputs), aData)

    state = DREAM.State(iNumChains, grid)
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 DREAM.Posterior(grid, likely, prior = "uniform", typeForm = DREAM.typeLogform),
                 DREAM.Domain(grid),
                 state,
                 DREAM.IndependentUpdate("gaussian", 0.01),
                 DREAM.DifferentialUpdate(100),
                 typeForm = DREAM.typeLogform)

    aSamples = state.getHistory()

    low_scale = numpy.average(numpy.array(
        [aSamples[i,0] if aSamples[i,1] < aSamples[i,3] else aSamples[i,2]
         for i in range(aSamples.shape[0])]
    ))
    low_rate = numpy.average(numpy.array(
        [aSamples[i,1] if aSamples[i,1] < aSamples[i,3] else aSamples[i,3]
         for i in range(aSamples.shape[0])]
    ))
    high_scale = numpy.average(numpy.array(
        [aSamples[i,0] if aSamples[i,1] >= aSamples[i,3] else aSamples[i,2]
         for i in range(aSamples.shape[0])]
    ))
    high_rate = numpy.average(numpy.array(
        [aSamples[i,1] if aSamples[i,1] >= aSamples[i,3] else aSamples[i,3]
         for i in range(aSamples.shape[0])]
    ))

    print("Acceptance rate: {0:1.3f}".format(state.getAcceptanceRate()))
    print("Inferred values:")
    print(" low   rate:{0:13.6f}   error:{1:14.6e}".format(low_rate,
                                                     numpy.abs(low_rate - 1.0)))
    print(" low  scale:{0:13.6f}   error:{1:14.6e}\n".format(low_scale,
                                                     numpy.abs(low_scale - 1.0)))
    print(" high  rate:{0:13.6f}   error:{1:14.6e}".format(high_rate,
                                                     numpy.abs(high_rate - 3.0)))
    print(" high scale:{0:13.6f}   error:{1:14.6e}".format(high_scale,
                                                    numpy.abs(high_scale - 0.4)))

if __name__ == "__main__":
    example_04()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from ctypes import c_char_p, c_int, c_double, c_void_p, POINTER, cdll, create_string_buffer
import numpy as np
import sys

from TasmanianConfig import __path_libdream__
from TasmanianConfig import TasmanianInputError as InputError

pLibDTSG = cdll.LoadLibrary(__path_libdream__)

pLibDTSG.tsgGetNumOutputs.restype = c_int

pLibDTSG.tsgDeleteLikelihood.argtypes = [c_void_p]
pLibDTSG.tsgGetNumOutputs.argtypes = [c_void_p]
pLibDTSG.tsgGetLikelihood.argtypes = [c_void_p, c_int, POINTER(c_double), c_int, POINTER(c_double)]

pLibDTSG.tsgMakeLikelihoodGaussIsotropic.restype = c_void_p
pLibDTSG.tsgMakeLikelihoodGaussIsotropic.argtypes = [c_int, c_double, POINTER(c_double), c_int]

pLibDTSG.tsgMakeLikelihoodGaussAnisotropic.restype = c_void_p
pLibDTSG.tsgMakeLikelihoodGaussAnisotropic.argtypes = [c_int, POINTER(c_double), POINTER(c_double), c_int]

typeRegform = 0
typeLogform = 1

class TasmanianLikelihood(object):
    '''
    Generic wrapper to TasDREAM::TasmanianLikelihood derived class.
    '''
    def __init__(self):
        '''
        The derived classes will have a member pClassPntr, init to nullptr here.
        '''
        self.TasmanianLikelihood = True
        self.pClassPntr = c_void_p(None)

    def __del__(self):
        '''
        Delete the grid on exit
        '''
        if self.pClassPntr:
            pLibDTSG.tsgDeleteLikelihood(self.pClassPntr)

    def getNumOutputs(self):
        '''
        Returns the number of outputs set in the likelihood.
        '''
        return pLibDTSG.tsgGetNumOutputs(self.pClassPntr)

    def getLikelihood(self, typeForm, llfModel):
        '''
        Returns the likelihood for the model outputs.

        typeForm: indicates to use the regular or log-form
                use DREAM.typeRegform or DREAM.typeLogform
        llfModel: is a two dimensional numpy.ndarray with
                .shape[0] > 0
                .shape[1] == getNumOutputs()

        Returns a one dimensional ndarray with size .shape[0]

        '''
        iNumSamples = llfModel.shape[0]
        iNumOutputs = llfModel.shape[1]
        aResult = np.empty((iNumSamples,), np.float64)
        pLibDTSG.tsgGetLikelihood(self.pClassPntr, typeForm,
                                  np.ctypeslib.as_ctypes(llfModel.reshape((iNumSamples*iNumOutputs,))),
                                  iNumSamples, np.ctypeslib.as_ctypes(aResult))
        return aResult


class LikelihoodGaussIsotropic(TasmanianLikelihood):
    '''
    Wrapper around TasDREAM::LikelihoodGaussIsotropic
    '''
    def __init__(self, fVariance, lfData, iNumSamples = 1):
        '''
        Make a new likelihood with the given data.

        fVariance:   float64 is the variance
        lfData:      one dimensional ndarray with the data
        iNumSamples: number of samples used for the data
        '''
        super(LikelihoodGaussIsotropic, self).__init__()
        iNumOutputs = len(lfData)
        self.pClassPntr = pLibDTSG.tsgMakeLikelihoodGaussIsotropic(
            iNumOutputs, fVariance, np.ctypeslib.as_ctypes(lfData), iNumSamples)


class LikelihoodGaussAnisotropic(TasmanianLikelihood):
    '''
    Wrapper around TasDREAM::LikelihoodGaussIsotropic
    '''
    def __init__(self, lfVariance, lfData, iNumSamples = 1):
        '''
        Make a new likelihood with the given data.

        lfVariance:  one dimensional ndarray with the data variance
        lfData:      one dimensional ndarray with the data
        iNumSamples: number of samples used for the data
        '''
        super(LikelihoodGaussAnisotropic, self).__init__()
        if (len(lfVariance) != len(lfData)):
            raise InputError("lfVariance", "Mismatch in size of variance and data vectors.")
        iNumOutputs = len(lfData)
        self.pClassPntr = pLibDTSG.tsgMakeLikelihoodGaussAnisotropic(
            iNumOutputs, np.ctypeslib.as_ctypes(lfVariance), np.ctypeslib.as_ctypes(lfData), iNumSamples)



########################################################################
# Pre-configured for GNU Make.
# The information here will be omnipresent throughout the testing
# environment, which will reduce the need for configuring.
########################################################################

bEnableSyncTests = True

sLibPath = "./libtasmaniansparsegrid.so"

iGPUID = -1 # not used by the GNU Make

bHasBlas = False
bHasCuBlas = False
bHasCuda = False

bUsingMSVC = False

sGaussPattersonTableFile = "GaussPattersonRule.table"



########################################################################
# This file will be configured by CMake (or pre-configured for GNU Make)
# The information here will be omnipresent throughout the testing
# environment, which will reduce the need for configuring.
########################################################################

bEnableSyncTests = True

sLibPath = "@Tasmanian_libsparsegrid_path@"

iGPUID = @Tasmanian_TESTS_GPU_ID@

bHasBlas = ("@Tasmanian_ENABLE_BLAS@" == "ON")
bHasCuBlas = ("@Tasmanian_ENABLE_CUDA@" == "ON")
bHasCuda = ("@Tasmanian_ENABLE_CUDA@" == "ON")

bUsingMSVC = ("@CMAKE_CXX_COMPILER_ID@" == "MSVC")

sGaussPattersonTableFile = "@CMAKE_CURRENT_BINARY_DIR@/GaussPattersonRule.table"



import unittest
import Tasmanian
DREAM = Tasmanian.DREAM
import os
import numpy as np

import testCommon

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Tests for the Tasmanian DREAM python bindings module
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkGaussian3D(self):
        iNumDimensions = 3
        iNumSamples = 1000
        iNumChains = 20
        iNumIterations = int(iNumSamples / iNumChains) + 2
        iNumBurnup = iNumIterations

        state = DREAM.State(iNumChains, iNumDimensions)
        state.setState(DREAM.genGaussianSamples([2.0, 2.0, 2.0], [3.0, 3.0, 3.0], iNumChains, DREAM.RandomGenerator("minstd_rand", 42)))

        DREAM.Sample(iNumBurnup, iNumSamples,
                     lambda x : np.exp( -0.5 * np.sum( (x - 2.0)**2, 1 ) / 9.0 ), # Gaussian mean 2.0 variance 9.0
                     DREAM.Domain("unbounded"),
                     state,
                     DREAM.IndependentUpdate("gaussian", 3.0),
                     DREAM.DifferentialUpdate(0),
                     DREAM.RandomGenerator("minstd_rand", 42))

        aMean, aVar = state.getHistoryMeanVariance()
        self.assertTrue(np.max(np.abs(aMean - 2.0)) < 0.2, "Failed the mean test for Gaussian 3D")
        self.assertTrue(np.max(np.abs(aVar - 9.0)) < 1.0, "Failed the variance test for Gaussian 3D")

        state = DREAM.State(iNumChains, iNumDimensions)
        state.setState(DREAM.genGaussianSamples([1.5, 2.0, 2.5], [0.5, 1.0, 2.0], iNumChains, DREAM.RandomGenerator("minstd_rand", 42)))

        likely = DREAM.LikelihoodGaussAnisotropic(np.array([0.25, 1.0, 4.0]), np.array([1.5, 2.0, 2.5]), 1)

        DREAM.Sample(iNumBurnup, iNumSamples,
                     lambda x : likely.getLikelihood(DREAM.typeLogform, x), # Gaussian mean 2.0 variance 9.0
                     DREAM.Domain("unbounded"),
                     state,
                     DREAM.IndependentUpdate("gaussian", 3.0),
                     DREAM.DifferentialUpdate(97),
                     DREAM.RandomGenerator("minstd_rand", 72),
                     DREAM.typeLogform)

        aMean, aVar = state.getHistoryMeanVariance()
        self.assertTrue(np.max(np.abs(aMean - np.array([1.5, 2.0, 2.5]))) < 0.4, "Failed the mean test for anisotropic Gaussian 3D")
        self.assertTrue(np.max(np.abs(aVar - np.array([0.25, 1.0, 4.0]))) < 1.0, "Failed the variance test for anisotropic Gaussian 3D")

    def checkGrid(self):
        iNumDimensions = 2
        iNumSamples = 1000
        iNumChains = 20
        iNumIterations = int(iNumSamples / iNumChains) + 2
        iNumBurnup = iNumIterations

        grid = Tasmanian.SparseGrid()
        grid.makeSequenceGrid(2, 1, 24, "level", "rleja")
        Tasmanian.loadNeededPoints(lambda x, tid : np.ones((1,)) * np.exp( -0.5 * np.sum((x - 0.3)**2) / 0.01 ), grid, 1)

        state = DREAM.State(iNumChains, grid)
        state.setState(DREAM.genGaussianSamples([0.3, 0.3], [0.1, 0.1], iNumChains, DREAM.RandomGenerator("minstd_rand", 55)))

        DREAM.Sample(iNumBurnup, iNumSamples,
                     DREAM.PosteriorEmbeddedLikelihood(grid, "uniform"),
                     DREAM.Domain("hypercube", np.array((0.0, 0.0)), np.array((1.0, 1.0))),
                     state,
                     DREAM.IndependentUpdate("uniform", 0.1),
                     DREAM.DifferentialUpdate(98),
                     DREAM.RandomGenerator("minstd_rand", 33))

        aMean, aVar = state.getHistoryMeanVariance()
        self.assertTrue(np.max(np.abs(aMean - np.array([0.3, 0.3]))) < 0.05, "Failed the mean test for grid likelihood")
        self.assertTrue(np.max(np.abs(aVar - np.array([0.01, 0.01]))) < 0.005, "Failed the variance test for grid likelihood")

    def checkModel(self):
        iNumDimensions = 2
        iNumSamples = 1000
        iNumChains = 20
        iNumIterations = int(iNumSamples / iNumChains) + 2
        iNumBurnup = iNumIterations

        likely = DREAM.LikelihoodGaussIsotropic(0.01, np.array([1.0, 2.0]))

        state = DREAM.State(iNumChains, iNumDimensions)
        state.setState(DREAM.genUniformSamples([-1.0, -1.0], [1.0, 1.0], iNumChains, DREAM.RandomGenerator("minstd_rand", 67)))

        DREAM.Sample(iNumBurnup, iNumSamples,
                     DREAM.Posterior(lambda x : np.matmul(np.cos(x), np.array([[1.0, 0.0], [0.0, 2.0]])),
                                     likely, "uniform", DREAM.typeLogform),
                     DREAM.Domain("hypercube", np.array((0.0, 0.0)), np.array((1.0, 1.0))),
                     state,
                     DREAM.IndependentUpdate("uniform", 0.1),
                     DREAM.DifferentialUpdate(98),
                     DREAM.RandomGenerator("minstd_rand", 88),
                     DREAM.typeLogform)

        aMode = state.getApproximateMode()
        self.assertTrue(np.max(np.abs(aMode)) < 0.01, "Failed the mode test for custom model and Tasmanian likelihood")

    def performDreamTests(self):
        self.checkGaussian3D() # analytic model/likelihood
        self.checkGrid()
        self.checkModel() # custom model

if __name__ == "__main__":
    tester = TestTasClass()
    tester.performDreamTests()



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from TasmanianDreamLikely import *
from TasmanianDreamSampler import *
from TasmanianDreamState import DreamState as State



#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from Tasmanian import DREAM
import numpy

def example_05():
    print("\n---------------------------------------------------------------------------------------------------\n")
    print("EXAMPLE 5: infer the frequency and magnitude of two signals from noisy data")
    print("           the model has 5 parameters: f(x_1 ... x_5) = sum x_k sin(k * pi * t)")
    print("           data = 2.0 * sin(2 * pi * t) + sin(4 * pi * t) + noise")
    print("           t in [0, 1], t is discretized using 64 equidistant nodes")
    print("           using two different likelihood functions, corresponding to l-2 and l-1 norms")
    print("           looking for the mode of the posterior, i.e., the optimal fit to the data\n")

    iNumDimensions = 5
    iNumChains = 50
    iNumBurnupIterations  = 1000
    iNumCollectIterations = 1000
    fScaleT = 1.0 / 64.0 # scale in T

    def model(aX):
        # the model assumes 2D array aX
        t = numpy.linspace(1.0 / 128.0, 1.0 - 1.0 / 128.0, 64)
        aSinMatrix = numpy.column_stack([numpy.sin(float(k) * numpy.pi * t) for k in range(1, 6)])
        return numpy.matmul(aSinMatrix, aX.transpose()).transpose()

    aNominalWeights = numpy.array([0.0, 2.0, 0.0, 1.0, 0.0])
    aData = model(aNominalWeights).reshape((64,))
    aData = aData + DREAM.genUniformSamples([-fScaleT for i in range(64)],
                                            [fScaleT for i in range(64)], 1).reshape((64,))

    likely = DREAM.LikelihoodGaussIsotropic(fScaleT, aData)

    domain_lower = numpy.array([0.0 for i in range(iNumDimensions)])
    domain_upper = numpy.array([3.0 for i in range(iNumDimensions)])

    state = DREAM.State(iNumChains, iNumDimensions)
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 DREAM.Posterior(model, likely, prior = "uniform", typeForm = DREAM.typeLogform),
                 DREAM.Domain("hypercube", domain_lower, domain_upper),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.0),
                 DREAM.DifferentialUpdate(100),
                 typeForm = DREAM.typeLogform)

    aSolution = state.getApproximateMode()

    sResult = ""
    sErrors = ""
    for i in range(iNumDimensions):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        sErrors = "{0:1s}{1:13.5e}".format(sErrors, numpy.abs(aSolution[i] - aNominalWeights[i]))
    print("Using Gaussian likelihood, the computed solution is:")
    print(" computed: {0:1s}".format(sResult))
    print("    error: {0:1s}".format(sErrors))

    ###############################################################################################
    # repeat the example, but change the likelihood
    ###############################################################################################
    def l1likelihood(aX, aData):
        # computes the likelihood, the 32 comes from the discretization in t
        aY = model(aX)
        aResult = numpy.empty((aY.shape[0],), numpy.float64)
        for i in range(aY.shape[0]):
            aResult[i] = -32.0 * numpy.sum(numpy.abs(aY[i,:] - aData))
        return aResult

    state = DREAM.State(iNumChains, iNumDimensions) # reset the state
    state.setState(DREAM.genUniformSamples(domain_lower, domain_upper, state.getNumChains()))

    DREAM.Sample(iNumBurnupIterations, iNumCollectIterations,
                 DREAM.PosteriorEmbeddedLikelihood(lambda x : l1likelihood(x, aData),
                                                   prior = "uniform",
                                                   typeForm = DREAM.typeLogform),
                 DREAM.Domain("hypercube", domain_lower, domain_upper),
                 state,
                 DREAM.IndependentUpdate("uniform", 0.0),
                 DREAM.DifferentialUpdate(100),
                 typeForm = DREAM.typeLogform)

    aSolution = state.getApproximateMode()

    sResult = ""
    sErrors = ""
    for i in range(iNumDimensions):
        sResult = "{0:1s}{1:13.5f}".format(sResult, aSolution[i])
        sErrors = "{0:1s}{1:13.5e}".format(sErrors, numpy.abs(aSolution[i] - aNominalWeights[i]))
    print("\nUsing l-1 likelihood, the computed solution is:")
    print(" computed: {0:1s}".format(sResult))
    print("    error: {0:1s}".format(sErrors))

if __name__ == "__main__":
    example_05()



import unittest
import TasmanianSG
import numpy as np

import testCommon # needed to compare grids after merge refinement

ttc = testCommon.TestTasCommon()

class TestTasClass(unittest.TestCase):
    '''
    Test methods for directly setting hierarchical matrices and coeffs,
    useful to build grids from unstructured/random data.
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def checkAgainstKnown(self):
        '''
        Check hierarchical functions against known (pen-and-paper)
        solutions.
        '''
        grid = TasmanianSG.TasmanianSparseGrid()

        # evalHierarchicalBasis (all done in batch), dense mode
        grid.makeLocalPolynomialGrid(2, 2, 1, 1, 'localp')
        aT = np.empty([0,2], np.float64)
        np.testing.assert_equal(aT, grid.getHierarchicalCoefficients(), "coeff mismatch", True)
        grid.makeLocalPolynomialGrid(2, 0, 1, 1, 'localp')
        aT = np.empty([0,0], np.float64)

        np.testing.assert_equal(aT, grid.getHierarchicalCoefficients(), "coeff mismatch", True)
        grid.makeLocalPolynomialGrid(2, 2, 1, 1, 'localp')
        aPoints = grid.getPoints()
        aV = np.column_stack([np.exp(2.0 * aPoints[:,0] + aPoints[:,1]), np.sin(3.0 * aPoints[:,0] + aPoints[:,1])])
        grid.loadNeededPoints(aV)
        aS = grid.getHierarchicalCoefficients()
        aT = np.array([[1.0, 0.0], [np.exp(-1.0)-1.0, np.sin(-1.0)], [np.exp(1.0)-1.0, np.sin(1.0)], [np.exp(-2.0)-1.0, np.sin(-3.0)], [np.exp(2.0)-1.0, np.sin(3.0)]])
        np.testing.assert_almost_equal(aS, aT, 14, "Surplusses equal", True)

        grid.makeGlobalGrid(3, 1, 4, 'level', 'fejer2')
        aPoints = grid.getPoints()
        aVan = grid.evaluateHierarchicalFunctions(aPoints)
        aIdentity = np.zeros([grid.getNumPoints(), grid.getNumPoints()], np.float64)
        for iI in range(grid.getNumPoints()):
            aIdentity[iI,iI] = 1.0
        np.testing.assert_almost_equal(aVan, aIdentity, 14, "evaluateHierarchicalFunctions", True)

        grid.makeGlobalGrid(2, 1, 1, 'level', 'clenshaw-curtis')
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])
        f0 = (1.0 - aPoints[:,0]**2) + (1.0 - aPoints[:,1]**2) - 1.0
        f1 = 0.5 * aPoints[:,1] * (aPoints[:,1] - 1.0)
        f2 = 0.5 * aPoints[:,1] * (aPoints[:,1] + 1.0)
        f3 = 0.5 * aPoints[:,0] * (aPoints[:,0] - 1.0)
        f4 = 0.5 * aPoints[:,0] * (aPoints[:,0] + 1.0)
        aResult = np.column_stack([f0, f1, f2, f3, f4])
        aVan = grid.evaluateHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aVan, aResult, 14, "evaluateHierarchicalFunctions", True)

        grid.makeSequenceGrid(2, 1, 2, 'level', 'leja')
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])
        f0 = np.ones([aPoints.shape[0],])
        f1 = aPoints[:,1]
        f2 = 0.5 * aPoints[:,1] * (aPoints[:,1] - 1.0)
        f3 = aPoints[:,0]
        f4 = aPoints[:,0] * aPoints[:,1]
        f5 = 0.5 * aPoints[:,0] * (aPoints[:,0] - 1.0)
        aResult = np.column_stack([f0, f1, f2, f3, f4, f5])
        aVan = grid.evaluateHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aVan, aResult, 14, "evaluateHierarchicalFunctions", True)

        grid.makeLocalPolynomialGrid(2, 1, 1, 1, 'localp')
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])

        f0 = np.ones([aPoints.shape[0],])
        f1 = -aPoints[:,1]
        f2 = np.copy(aPoints[:,1])
        f3 = -aPoints[:,0]
        f4 = np.copy(aPoints[:,0])
        for iI in range(aPoints.shape[0]):
            if (aPoints[iI,1] > 0.0): f1[iI] = 0.0
            if (aPoints[iI,1] < 0.0): f2[iI] = 0.0
            if (aPoints[iI,0] > 0.0): f3[iI] = 0.0
            if (aPoints[iI,0] < 0.0): f4[iI] = 0.0
        aResult = np.column_stack([f0, f1, f2, f3, f4])
        aVan = grid.evaluateHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aVan, aResult, 14, "evaluateHierarchicalFunctions", True)

        grid.makeFourierGrid(2, 1, 1, "level")
        aPoints = np.array([[0.25, 0.5]]);
        aResult = np.array([[1.0 + 0.0 * 1j, -1.0 + 0.0 * 1j, -1.0 + 0.0 * 1j, 0.0 - 1.0 * 1j, 0.0 + 1.0 * 1j]])
        aVan = grid.evaluateHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aVan, aResult, 14, "evaluateHierarchicalFunctions", True)

        # sparse hierarchical functions
        pSparse = TasmanianSG.TasmanianSimpleSparseMatrix()
        np.testing.assert_almost_equal(np.empty([0,0], np.float64), pSparse.getDenseForm(), 14, "TasmanianSimpleSparseMatrix.getDense()", True)

        grid.makeLocalPolynomialGrid(2, 1, 4, 1, 'localp')
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])
        aDense = grid.evaluateHierarchicalFunctions(aPoints)
        pSparse = grid.evaluateSparseHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aDense, pSparse.getDenseForm(), 14, "evaluateSparseHierarchicalFunctions", True)

        grid.makeWaveletGrid(2, 1, 4, 1)
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])
        aDense = grid.evaluateHierarchicalFunctions(aPoints)
        pSparse = grid.evaluateSparseHierarchicalFunctions(aPoints)
        np.testing.assert_almost_equal(aDense, pSparse.getDenseForm(), 14, "evaluateSparseHierarchicalFunctions", True)

    def checkSetCoeffsMergeRefine(self):
        '''
        Set the coefficients and use mergeRefinement()
        '''
        aPoints = np.array([[0.33, 0.25], [-0.27, 0.39], [0.97, -0.76], [-0.44, 0.21], [-0.813, 0.03], [-0.666, 0.666]])
        # set hierarchical coeffs/merge refinement
        gridA = TasmanianSG.TasmanianSparseGrid()
        gridB = TasmanianSG.TasmanianSparseGrid()
        gridA.makeGlobalGrid(2, 1, 4, 'level', 'chebyshev')
        gridB.makeGlobalGrid(2, 1, 4, 'level', 'chebyshev')
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

        gridA = TasmanianSG.TasmanianSparseGrid()
        gridB = TasmanianSG.TasmanianSparseGrid()
        gridA.makeGlobalGrid(2, 1, 4, 'level', 'fejer2')
        gridB.makeGlobalGrid(2, 1, 4, 'level', 'fejer2')
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)
        gridA.setAnisotropicRefinement('iptotal', 10, 0)
        gridB.setAnisotropicRefinement('iptotal', 10, 0)
        ttc.loadExpN2(gridA)
        gridB.mergeRefinement()
        self.assertEqual(gridA.getNumPoints(), gridB.getNumPoints(), 'different number of points after merge refinement')
        aRes = gridB.evaluateBatch(aPoints)
        np.testing.assert_almost_equal(aRes, np.zeros(aRes.shape), 14, "not zero after merged refinement", True)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

        gridA.makeSequenceGrid(2, 1, 5, 'ipcurved', 'min-delta')
        gridB.makeSequenceGrid(2, 1, 5, 'ipcurved', 'min-delta')
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)
        gridA.setAnisotropicRefinement('iptotal', 30, 0)
        gridB.setAnisotropicRefinement('iptotal', 30, 0)
        ttc.loadExpN2(gridA)
        gridB.mergeRefinement()
        self.assertEqual(gridA.getNumPoints(), gridB.getNumPoints(), 'different number of points after merge refinement')
        aRes = gridB.evaluateBatch(aPoints)
        np.testing.assert_almost_equal(aRes, np.zeros(aRes.shape), 14, "not zero after merged refinement", True)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

        gridA.makeLocalPolynomialGrid(2, 1, 4, 2, 'semi-localp')
        gridB.makeLocalPolynomialGrid(2, 1, 4, 2, 'semi-localp')
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)
        gridA.setSurplusRefinement(1.E-4, 0, 'classic')
        gridB.setSurplusRefinement(1.E-4, 0, 'classic')
        ttc.loadExpN2(gridA)
        gridB.mergeRefinement()
        self.assertEqual(gridA.getNumPoints(), gridB.getNumPoints(), 'different number of points after merge refinement')
        aRes = gridB.evaluateBatch(aPoints)
        np.testing.assert_almost_equal(aRes, np.zeros(aRes.shape), 14, "not zero after merged refinement", True)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

        gridA.makeWaveletGrid(2, 1, 2, 1)
        gridB.makeWaveletGrid(2, 1, 2, 1)
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)
        gridA.setSurplusRefinement(1.E-3, 0, 'classic')
        gridB.setSurplusRefinement(1.E-3, 0, 'classic')
        ttc.loadExpN2(gridA)
        gridB.mergeRefinement()
        self.assertEqual(gridA.getNumPoints(), gridB.getNumPoints(), 'different number of points after merge refinement')
        aRes = gridB.evaluateBatch(aPoints)
        np.testing.assert_almost_equal(aRes, np.zeros(aRes.shape), 14, "not zero after merged refinement", True)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

        gridA.makeFourierGrid(2, 1, 3, 'level')
        gridB.makeFourierGrid(2, 1, 3, 'level')
        ttc.loadExpN2(gridA)
        gridB.setHierarchicalCoefficients(gridA.getHierarchicalCoefficients())
        ttc.compareGrids(gridA, gridB)

    def checkIntegrals(self):
        grid = TasmanianSG.TasmanianSparseGrid()

        lTests = ['grid.makeGlobalGrid(2, 1, 4, "level", "clenshaw-curtis")',
                  'grid.makeSequenceGrid(2, 1, 4, "level", "rleja")',
                  'grid.makeFourierGrid(2, 1, 3, "level")',
                  'grid.makeLocalPolynomialGrid(2, 1, 4)',
                  'grid.makeWaveletGrid(2, 1, 2)']

        for t in lTests:
            exec(t)
            grid.setDomainTransform(np.array([[-3.0, 5.0], [-4.0, 2.0]]))
            ttc.loadExpN2(grid)
            aIntegrals = grid.integrateHierarchicalFunctions()
            aSurps = grid.getHierarchicalCoefficients()
            fResult = np.real(np.sum(aIntegrals * aSurps.reshape((aSurps.shape[0],))))
            fExpected = grid.integrate()
            np.testing.assert_almost_equal(fResult, fExpected[0], 12, 'integrals different by too much')

        # check support
        grid.makeLocalPolynomialGrid(1, 1, 3)
        aSup = grid.getHierarchicalSupport()
        aResult = (1.0, 1.0, 1.0, 0.5, 0.5, 0.25, 0.25, 0.25, 0.25)
        np.testing.assert_almost_equal(aSup, np.array(aResult).reshape((9, 1)), 12, 'support different by too much')

    def checkValues(self):
        lTests = ['grid.makeGlobalGrid(2, 2, 3, "level", "clenshaw-curtis")',
                  'grid.makeSequenceGrid(2, 2, 3, "level", "rleja")',
                  'grid.makeFourierGrid(2, 2, 2, "level")',
                  'grid.makeLocalPolynomialGrid(2, 2, 2)',
                  'grid.makeWaveletGrid(2, 2, 2)']

        for t in lTests:
            grid = TasmanianSG.TasmanianSparseGrid()
            aResult = grid.getLoadedValues()
            self.assertTrue(len(aResult) == 0, "failed to return a empty array")
            exec(t)
            iNP = grid.getNumNeeded()
            iOuts = 2
            aReferece = np.array([float(i) for i in range(iNP * iOuts)]).reshape(iNP, iOuts)
            grid.loadNeededPoints(aReferece)
            aResult = grid.getLoadedValues()
            np.testing.assert_almost_equal(aResult, aReferece, 14, 'values in getLoadedValues() differ by too much')

    def performUnstructuredDataTests(self):
        self.checkAgainstKnown()
        self.checkSetCoeffsMergeRefine()
        self.checkIntegrals()
        self.checkValues()



#!@Tasmanian_string_python_hashbang@

##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

# The sys.path.append() command is necessary only if Tasmanian is
# not included in the system PYTHONPATH
# If PYTHONPATH is set (e.g., source TasmanianENVsetup.sh) you can jump
# straight to import Tasmanian
import sys
@Tasmanian_python_example_import@
import Tasmanian

import example_sparse_grids_01
import example_sparse_grids_02
import example_sparse_grids_03
import example_sparse_grids_04
import example_sparse_grids_05
import example_sparse_grids_06
import example_sparse_grids_07
import example_sparse_grids_08
import example_sparse_grids_09
import example_sparse_grids_10

if __name__ == "__main__":
    example_sparse_grids_01.example_01()
    example_sparse_grids_02.example_02()
    example_sparse_grids_03.example_03()
    example_sparse_grids_04.example_04()
    example_sparse_grids_05.example_05()
    example_sparse_grids_06.example_06()
    example_sparse_grids_07.example_07()
    example_sparse_grids_08.example_08()
    example_sparse_grids_09.example_09()
    example_sparse_grids_10.example_10()
    print("")


exit(0)

import numpy as np

# imports specifically needed by the examples
import math
from random import uniform
from datetime import datetime

bFastTests = False
if (len(sys.argv) > 1):
    bFastTests = True

print("Tasmanian version: {0:s}".format(Tasmanian.__version__))
print("Tasmanian license: {0:s}".format(Tasmanian.__license__))

grid = Tasmanian.SparseGrid()

iDim = 2
iLevel = 6

# EXAMPLE 1: integrate: f(x,y) = exp(-x^2) * cos(y) over [-1,1] x [-1,1]
# using classical Smolyak grid with Clenshaw-Curtis points and weights
grid = Tasmanian.SparseGrid()

print("\n-------------------------------------------------------------------------------------------------")
print("Example 1:  integrate f(x,y) = exp(-x^2) * cos(y), using clenshaw-curtis level nodes")

iDim = 2
iLevel = 6

grid.makeGlobalGrid(iDim, 0, iLevel, "level", "clenshaw-curtis")
aPoints = grid.getPoints()
aWeights = grid.getQuadratureWeights()

fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(aPoints.shape[0])])
fError = math.fabs(fIntegral - 2.513723354063905e+00)

print("      at level: {0:d}".format(iLevel))
print("      the grid has: {0:d}".format(aPoints.shape[0]))
print("      integral: {0:1.16e}".format(fIntegral))
print("         error: {0:1.16e}\n".format(fError))

iLevel = 7

grid.makeGlobalGrid(iDim, 0, iLevel, "level", "clenshaw-curtis")
aPoints = grid.getPoints()
aWeights = grid.getQuadratureWeights()

fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(aPoints.shape[0])])
fError = math.fabs(fIntegral - 2.513723354063905e+00)

print("      at level: {0:d}".format(iLevel))
print("      the grid has: {0:d}".format(aPoints.shape[0]))
print("      integral: {0:1.16e}".format(fIntegral))
print("         error: {0:1.16e}".format(fError))

#
# EXAMPLE 2: integrate: f(x,y) = exp(-x^2) * cos(y) over (x,y) in [-5,5] x [-2,3]
# using Gauss-Patterson rules chosen to integrate exactly polynomials of
# total degree up to degree specified by prec

print("\n-------------------------------------------------------------------------------------------------")
print("Example 2: integrate f(x,y) = exp(-x^2) * cos(y) over [-5,5] x [-2,3] using  Gauss-Patterson nodes")

iDim = 2
iPrec = 20

grid.makeGlobalGrid(iDim, 0, iPrec, "qptotal", "gauss-patterson")

grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 3.0]]))

aPoints = grid.getPoints()
aWeights = grid.getQuadratureWeights()

fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(aPoints.shape[0])])
fError = math.fabs(fIntegral - 1.861816427518323e+00)

print("      at precision: {0:d}".format(iPrec))
print("      the grid has: {0:d}".format(aPoints.shape[0]))
print("      integral: {0:1.16e}".format(fIntegral))
print("         error: {0:1.16e}".format(fError))

iPrec = 40

grid.makeGlobalGrid(iDim, 0, iPrec, "qptotal", "gauss-patterson")

grid.setDomainTransform(np.array([[-5.0, 5.0], [-2.0, 3.0]]))

aPoints = grid.getPoints()
aWeights = grid.getQuadratureWeights()

fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(aPoints.shape[0])])
fError = math.fabs(fIntegral - 1.861816427518323e+00)

print("      at precision: {0:d}".format(iPrec))
print("      the grid has: {0:d}".format(aPoints.shape[0]))
print("      integral: {0:1.16e}".format(fIntegral))
print("         error: {0:1.16e}".format(fError))

#
# EXAMPLE 3: integrate: f(x,y) = exp(-x^2) * cos(y) over (x,y) in [-5,5] x [-2,3]
# using different rules

print("\n-------------------------------------------------------------------------------------------------")
print("Example 3: integrate f(x,y) = exp(-x^2) * cos(y) over [-5,5] x [-2,3] using different rules\n")

iDim = 2
fExact = 1.861816427518323e+00
mRange = np.array([[-5.0, 5.0], [-2.0, 3.0]])

print("           Clenshaw-Curtis      Gauss-Legendre     Gauss-Patterson")
print(" precision    points     error    points     error    points     error")

liPrecisions = [9, 13, 17, 21, 25, 29]
for iPrec in liPrecisions:
    grid1 = Tasmanian.TasmanianSparseGrid()
    grid1.makeGlobalGrid(iDim, 0, iPrec, "qptotal", "clenshaw-curtis")
    grid2 = Tasmanian.TasmanianSparseGrid()
    grid2.makeGlobalGrid(iDim, 0, iPrec, "qptotal", "gauss-legendre")
    grid3 = Tasmanian.TasmanianSparseGrid()
    grid3.makeGlobalGrid(iDim, 0, iPrec, "qptotal", "gauss-patterson")

    grid1.setDomainTransform(mRange)
    grid2.setDomainTransform(mRange)
    grid3.setDomainTransform(mRange)

    aPoints = grid1.getPoints()
    aWeights = grid1.getQuadratureWeights()
    iNumP1 = aPoints.shape[0]
    fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(iNumP1)])
    fError1 = math.fabs(fIntegral - fExact)

    aPoints = grid2.getPoints()
    aWeights = grid2.getQuadratureWeights()
    iNumP2 = aPoints.shape[0]
    fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(iNumP2)])
    fError2 = math.fabs(fIntegral - fExact)

    aPoints = grid3.getPoints()
    aWeights = grid3.getQuadratureWeights()
    iNumP3 = aPoints.shape[0]
    fIntegral = sum([aWeights[i] * math.exp(-aPoints[i][0]*aPoints[i][0]) * math.cos(aPoints[i][1]) for i in range(iNumP3)])
    fError3 = math.fabs(fIntegral - fExact)

    print("    {0:2d}       {1:3d}  {2:1.2e}      {3:4d}  {4:1.2e}       {5:3d}  {6:1.2e}".format(iPrec, iNumP1, fError1, iNumP2, fError2, iNumP3, fError3))

#
# EXAMPLE 4: interpolate: f(x,y) = exp(-x^2) * cos(y)
# with a rule that exactly interpolates polynomials of total degree up to
# degree specified by prec

print("\n-------------------------------------------------------------------------------------------------")
print("Example 4: interpolate f(x,y) = exp(-x^2) * cos(y), using clenshaw-curtis iptotal rule\n")

iDim = 2
iOut = 1
iPrec = 10
aX = np.array([0.3, 0.7])
fExact = math.exp(-aX[0]*aX[0]) * math.cos(aX[1])

grid.makeGlobalGrid(iDim, iOut, iPrec, "iptotal", "clenshaw-curtis")

aPoints = grid.getPoints()

aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI][0] = math.exp(- aPoints[iI][0] * aPoints[iI][0]) * math.cos(aPoints[iI][1])

grid.loadNeededPoints(aVals)

aY = grid.evaluate(aX)

print("  using polynomials of total degree up to: {0:1d}".format(iPrec))
print("                 the grid has: {0:1d} points".format(aPoints.shape[0]))
print("         interpolant at (0.3,0.7): {0:1.16e}".format(aY[0]))
print("                    error: {0:1.16e}\n".format(math.fabs(aY[0] - fExact)))

iPrec = 12
grid.makeGlobalGrid(iDim, iOut, iPrec, "iptotal", "clenshaw-curtis")

aPoints = grid.getPoints()

aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI][0] = math.exp(- aPoints[iI][0] * aPoints[iI][0]) * math.cos(aPoints[iI][1])

grid.loadNeededPoints(aVals)

aY = grid.evaluate(aX)

print("  using polynomials of total degree up to: {0:1d}".format(iPrec))
print("                 the grid has: {0:1d} points".format(aPoints.shape[0]))
print("         interpolant at (0.3,0.7): {0:1.16e}".format(aY[0]))
print("                    error: {0:1.16e}".format(math.fabs(aY[0] - fExact)))

if (bFastTests):
    exit(0)

# random points to be used for error checking
aPnts = np.empty([1000, 4])
for iI in range(1000):
    for iJ in range(4):
        aPnts[iI][iJ] = uniform(-1.0, 1.0)

#
# EXAMPLE 5:
# interpolate: f(x1,x2,x3,x4) = exp(-x1^2) * cos(x2) * exp(-x3^2) * cos(x4)
# with Global and Sequence Leja rules

iDim = 4
iOut = 1
iPrec = 15

aTres = np.empty([1000,])
for iI in range(1000):
    aTres[iI] = math.exp(- aPnts[iI][0] ** 2) * math.cos(aPnts[iI][1]) * math.exp(- aPnts[iI][2] ** 2) * math.cos(aPnts[iI][3])

clock_start = datetime.now()
grid.makeGlobalGrid(iDim, iOut, iPrec, "iptotal", "leja")
clock_end = datetime.now()
stage1 = clock_end - clock_start

aPoints = grid.getPoints()

print("\n-------------------------------------------------------------------------------------------------")
print("Example 5: interpolate f(x1,x2,x3,x4) = exp(-x1^2) * cos(x2) * exp(-x3^2) * cos(x4)")
print("       comparign the performance of Global and Sequence grids with leja nodes")
print("       using polynomials of total degree up to: {0:1d}".format(iPrec))
print("                    grids have: {0:1d} points".format(aPoints.shape[0]))
print("       both grids are evaluated at 1000 random points \n")

aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1]) * math.exp(- aPoints[iI][2] ** 2) * math.cos(aPoints[iI][3])

clock_start = datetime.now()
grid.loadNeededPoints(aVals)
clock_end = datetime.now()
stage2 = clock_end - clock_start

clock_start = datetime.now()
aRes = grid.evaluateBatch(aPnts)
clock_end = datetime.now()
stage3 = clock_end - clock_start

clock_start = datetime.now()
grid.makeSequenceGrid(iDim, iOut, iPrec, "iptotal", "leja")
clock_end = datetime.now()
sstage1 = clock_end - clock_start

clock_start = datetime.now()
grid.loadNeededPoints(aVals)
clock_end = datetime.now()
sstage2 = clock_end - clock_start

clock_start = datetime.now()
aRes = grid.evaluateBatch(aPnts)
clock_end = datetime.now()
sstage3 = clock_end - clock_start

print("          Stage    Global Grid  Sequence Grid")
print("      make grid {0:14d} {1:14d}   microseconds".format(stage1.microseconds, sstage1.microseconds))
print("    load values {0:14d} {1:14d}   microseconds".format(stage2.microseconds, sstage2.microseconds))
print("       evaluate {0:14d} {1:14d}   microseconds".format(stage3.microseconds, sstage3.microseconds))

print("\n NOTE: due to the somewhat random nature of garbage collection in Python,")
print("       the actual timing above may be inacurate")

#
# EXAMPLE 6:
# interpolate: f(x,y) = exp(-x^2) * cos(y)
# using different refinement schemes

aPnts = aPnts[:,:2]
aTres = np.empty([1000,])
for iI in range(1000):
    aTres[iI] = math.exp(- aPnts[iI][0] ** 2) * math.cos(aPnts[iI][1])

iDim = 2
iOut = 1

grid1.makeGlobalGrid(iDim, iOut, 3, "iptotal", "leja")
grid2.makeGlobalGrid(iDim, iOut, 3, "iptotal", "leja")
grid3.makeSequenceGrid(iDim, iOut, 3, "iptotal", "leja")

aPoints = grid1.getPoints()
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1])

grid1.loadNeededPoints(aVals)
grid2.loadNeededPoints(aVals)
grid3.loadNeededPoints(aVals)

print("\n-------------------------------------------------------------------------------------------------")
print("Example 6: interpolate: f(x,y) = exp(-x^2) * cos(y)")
print("   using leja nodes and different refinement schemes")
print("   the error is estimated as the maximum from 1000 random points\n")

print("          Toral Degree          Curved         Surplus")
print(" iteration    points     error    points     error    points     error")

for iK in range(10):
    grid1.setAnisotropicRefinement("iptotal", 10, 0)
    aPoints = grid1.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1])
    grid1.loadNeededPoints(aVals)

    aRes = grid1.evaluateBatch(aPnts)
    fError1 = max(np.fabs(aRes[:,0] - aTres))

    grid2.setAnisotropicRefinement("ipcurved", 10, 0)
    aPoints = grid2.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1])
    grid2.loadNeededPoints(aVals)

    aRes = grid2.evaluateBatch(aPnts)
    fError2 = max(np.fabs(aRes[:,0] - aTres))

    grid3.setSurplusRefinement(1.E-10, 0)
    aPoints = grid3.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1])
    grid3.loadNeededPoints(aVals)

    aRes = grid3.evaluateBatch(aPnts)
    fError3 = max(np.fabs(aRes[:,0] - aTres))

    print(" {0:9d} {1:9d}  {2:1.2e} {3:9d}  {4:1.2e} {5:9d}  {6:1.2e}".format(iK+1, grid1.getNumPoints(), fError1, grid2.getNumPoints(), fError2, grid3.getNumPoints(), fError3))

#
# EXAMPLE 7:
# interpolate: f(x,y) = exp(-x^2) * cos(y)
# using localp and semilocalp grids

iDim = 2
iOut = 1
iDepth = 7

print("\n-------------------------------------------------------------------------------------------------")
print("Example 7: interpolate: f(x,y) = exp(-x^2) * cos(y)")
print("       using localp and semi-localp rules with depth {0:1d}".format(iDepth))
print("       the error is estimated as the maximum from 1000 random points\n")

grid.makeLocalPolynomialGrid(iDim, iOut, iDepth, 2, "localp")

aPoints = grid.getPoints()
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(- aPoints[iI][0] ** 2) * math.cos(aPoints[iI][1])
grid.loadNeededPoints(aVals)

aRes = grid.evaluateBatch(aPnts)
fError1 = max(np.fabs(aRes[:,0] - aTres))

grid.makeLocalPolynomialGrid(iDim, iOut, iDepth, 2, "semi-localp")

# localp and semi-localp grids have the same points, no need to recompute aVals
grid.loadNeededPoints(aVals)

aRes = grid.evaluateBatch(aPnts)
fError2 = max(np.fabs(aRes[:,0] - aTres))

print("     Number of points: {0:1d}".format(aPoints.shape[0]))
print("     Error for    localp: {0:1.4e}".format(fError1))
print("     Error for   semi-localp: {0:1.4e}".format(fError2))
print(" Note: semi-localp wins this competition because the function is very smooth")

#
# EXAMPLE 8:
# interpolate: f(x,y) = cos(0.5 * pi * x) * cos(0.5 * pi * y)
# using localp and semilocalp grids

aTres = np.empty([1000,])
for iI in range(1000):
    aTres[iI] = math.cos(0.5 * math.pi * aPnts[iI][0]) * math.cos(0.5 * math.pi * aPnts[iI][1])

iDim = 2
iOut = 1
iDepth = 7

print("\n-------------------------------------------------------------------------------------------------")
print("Example 8: interpolate f(x,y) = cos(0.5 * pi * x) * cos(0.5 * pi * y)")
print("       using localp and localp-zero rules with depth {0:1d}".format(iDepth))
print("       the error is estimated as the maximum from 1000 random points\n")

grid.makeLocalPolynomialGrid(iDim, iOut, iDepth, 2, "localp")

aPoints = grid.getPoints()
iNumP1 = aPoints.shape[0]
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.cos(0.5 * math.pi * aPoints[iI][0]) * math.cos(0.5 * math.pi * aPoints[iI][1])
grid.loadNeededPoints(aVals)

aRes = grid.evaluateBatch(aPnts)
fError1 = max(np.fabs(aRes[:,0] - aTres))

grid.makeLocalPolynomialGrid(iDim, iOut, iDepth-1, 2, "localp-zero")

aPoints = grid.getPoints()
iNumP2 = aPoints.shape[0]
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.cos(0.5 * math.pi * aPoints[iI][0]) * math.cos(0.5 * math.pi * aPoints[iI][1])
grid.loadNeededPoints(aVals)

aRes = grid.evaluateBatch(aPnts)
fError2 = max(np.fabs(aRes[:,0] - aTres))

print(" For localp    Number of points: {0:1d}   Error: {1:1.16e}".format(iNumP1, fError1))
print(" For localp-zero   Number of points: {0:1d}   Error: {1:1.16e}".format(iNumP2, fError2))
print(" Note: localp-zero wins this competition because the function is zero at the boundary")

#
# EXAMPLE 9:
# interpolate: f(x,y) = exp(-x) / (1 + 100 * exp(-10 * y))
# using different refinement schemes

aTres = np.empty([1000,])
for iI in range(1000):
    aTres[iI] = math.exp(-aPnts[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPnts[iI][1]))

iDim = 2
iOut = 1
iDepth = 2
fTol = 1.E-5

grid1.makeLocalPolynomialGrid(iDim, iOut, iDepth, -1, "localp")

aPoints = grid1.getPoints()
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
grid1.loadNeededPoints(aVals)

grid2.makeLocalPolynomialGrid(iDim, iOut, iDepth, -1, "localp")
grid2.loadNeededPoints(aVals)

print("\n-------------------------------------------------------------------------------------------------")
print("Example 9: interpolate f(x,y) = exp(-x) / (1 + 100 * exp(-10 * y))")
print("   the error is estimated as the maximum from 1000 random points")
print("   tolerance is set at 1.E-5 and maximal order polynomials are used\n")

print("               Classic         FDS")
print(" precision    points     error    points     error")

for iK in range(7):
    grid1.setSurplusRefinement(fTol, -1, "classic")
    aPoints = grid1.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
    grid1.loadNeededPoints(aVals)

    aRes = grid1.evaluateBatch(aPnts)
    fError1 = max(np.fabs(aRes[:,0] - aTres))

    grid2.setSurplusRefinement(fTol, -1, "fds")
    aPoints = grid2.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
    grid2.loadNeededPoints(aVals)

    aRes = grid2.evaluateBatch(aPnts)
    fError2 = max(np.fabs(aRes[:,0] - aTres))

    print(" {0:9d} {1:9d}  {2:1.2e} {3:9d}  {4:1.2e}".format(iK+1, grid1.getNumPoints(), fError1, grid2.getNumPoints(), fError2))

#
# EXAMPLE 10:
# interpolate: f(x,y) = exp(-x) / (1 + 100 * exp(-10 * y))
# using local polynomails and wavelets

grid1.makeLocalPolynomialGrid(iDim, iOut, iDepth=3, iOrder=1, sRule="localp")
aPoints = grid1.getPoints()
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
grid1.loadNeededPoints(aVals)

grid2.makeWaveletGrid(iDim, iOut, iDepth=1, iOrder=1)
aPoints = grid2.getPoints()
aVals = np.empty([aPoints.shape[0], 1])
for iI in range(aPoints.shape[0]):
    aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
grid2.loadNeededPoints(aVals)

print("\n-------------------------------------------------------------------------------------------------")
print("Example 10: interpolate f(x,y) = exp(-x) / (1 + 100 * exp(-10 * y))")
print("   the error is estimated as the maximum from 1000 random points")
print("   using local polynomials and wavelets\n")

print("           Polynomials        Wavelets")
print(" precision    points     error    points     error")

for iK in range(8):
    grid1.setSurplusRefinement(fTol, -1, "fds")
    aPoints = grid1.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
    grid1.loadNeededPoints(aVals)

    aRes = grid1.evaluateBatch(aPnts)
    fError1 = max(np.fabs(aRes[:,0] - aTres))

    grid2.setSurplusRefinement(fTol, -1, "fds")
    aPoints = grid2.getNeededPoints()
    aVals = np.empty([aPoints.shape[0], 1])
    for iI in range(aPoints.shape[0]):
        aVals[iI] = math.exp(-aPoints[iI][0]) / (1.0 + 100.0 * math.exp(-10.0 * aPoints[iI][1]))
    grid2.loadNeededPoints(aVals)

    aRes = grid2.evaluateBatch(aPnts)
    fError2 = max(np.fabs(aRes[:,0] - aTres))

    print(" {0:9d} {1:9d}  {2:1.2e} {3:9d}  {4:1.2e}".format(iK+1, grid1.getNumPoints(), fError1, grid2.getNumPoints(), fError2))


# random points to be used for error checking
aPnts = np.empty([1000, 3])
for iI in range(1000):
    for iJ in range(3):
        aPnts[iI][iJ] = uniform(-1.0, 1.0)


#
# EXAMPLE 11:
#
# interpolate: f(x,y,z) = 1/((1+4x^2)*(1+5y^2)*(1+6z^2))
# using classical and conformal transformation
#
aTres = 1.0 / ((1.0 + 4.0*aPnts[:,0]**2)*(1.0 + 5.0*aPnts[:,1]**2)*(1.0 + 6.0*aPnts[:,2]**2));

iDim = 3;
iPrec = 12;
iOuts = 1;
grid.makeGlobalGrid(iDim, iOuts, iPrec, 'iptotal', 'clenshaw-curtis');
aPoints = grid.getNeededPoints()
aVals = 1.0 / ((1.0 + 4.0*aPoints[:,0]**2)*(1.0 + 5.0*aPoints[:,1]**2)*(1.0 + 6.0*aPoints[:,2]**2));
grid.loadNeededPoints(aVals.reshape([aPoints.shape[0], 1]));
aRes = grid.evaluateBatch(aPnts);
fError1 = max(np.fabs(aRes[:,0] - aTres))
iNump1 = aPoints.shape[0]

grid.makeGlobalGrid(iDim, iOuts, iPrec, 'iptotal', 'clenshaw-curtis');
grid.setConformalTransformASIN(np.array([4, 4, 4]))
aPoints = grid.getNeededPoints()
aVals = 1.0 / ((1.0 + 4.0*aPoints[:,0]**2)*(1.0 + 5.0*aPoints[:,1]**2)*(1.0 + 6.0*aPoints[:,2]**2));
grid.loadNeededPoints(aVals.reshape([aPoints.shape[0], 1]));
aRes = grid.evaluateBatch(aPnts);
fError2 = max(np.fabs(aRes[:,0] - aTres))

grid.makeLocalPolynomialGrid(iDim, iOuts, iPrec-4, 2);
aPoints = grid.getNeededPoints()
aVals = 1.0 / ((1.0 + 4.0*aPoints[:,0]**2)*(1.0 + 5.0*aPoints[:,1]**2)*(1.0 + 6.0*aPoints[:,2]**2));
grid.loadNeededPoints(aVals.reshape([aPoints.shape[0], 1]));
aRes = grid.evaluateBatch(aPnts);
fError3 = max(np.fabs(aRes[:,0] - aTres))
iNump2 = aPoints.shape[0]

grid.makeLocalPolynomialGrid(iDim, iOuts, iPrec-4, 2);
grid.setConformalTransformASIN(np.array([4, 4, 4]))
aPoints = grid.getNeededPoints()
aVals = 1.0 / ((1.0 + 4.0*aPoints[:,0]**2)*(1.0 + 5.0*aPoints[:,1]**2)*(1.0 + 6.0*aPoints[:,2]**2));
grid.loadNeededPoints(aVals.reshape([aPoints.shape[0], 1]));
aRes = grid.evaluateBatch(aPnts);
fError4 = max(np.fabs(aRes[:,0] - aTres))
iNump2 = aPoints.shape[0]


print("\n----------------------------------------------------------------------------")
print("Example 11: interpolate f(x,y,z) = 1/((1+4x^2)*(1+5y^2)*(1+6z^2))")
print("            using conformal transformation")
print("            the error is estimated as the maximum from 1000 random points")
print("");
print("Grid Type    nodes     error regular   error conformal")
print("Global         {0:1d}         {1:1.3e}         {2:1.3e}".format(iNump1, fError1, fError2))
print("Localp        {0:1d}         {1:1.3e}         {2:1.3e}".format(iNump2, fError3, fError4))
print("")
print("Note: conformal maps address specific problems with the region of analyticity of a function")
print("      the map can accelerate or slow down convergence depending on the problem")

print("\n-------------------------------------------------------------------------------------------------\n")



##############################################################################################################################################################################
# Copyright (c) 2017, Miroslav Stoyanov
#
# This file is part of
# Toolkit for Adaptive Stochastic Modeling And Non-Intrusive ApproximatioN: TASMANIAN
#
# Redistribution and use in source and binary forms, with or without modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice, this list of conditions
#    and the following disclaimer in the documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors may be used to endorse
#    or promote products derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES,
# INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
# IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY,
# OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA,
# OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#
# UT-BATTELLE, LLC AND THE UNITED STATES GOVERNMENT MAKE NO REPRESENTATIONS AND DISCLAIM ALL WARRANTIES, BOTH EXPRESSED AND IMPLIED.
# THERE ARE NO EXPRESS OR IMPLIED WARRANTIES OF MERCHANTABILITY OR FITNESS FOR A PARTICULAR PURPOSE, OR THAT THE USE OF THE SOFTWARE WILL NOT INFRINGE ANY PATENT,
# COPYRIGHT, TRADEMARK, OR OTHER PROPRIETARY RIGHTS, OR THAT THE SOFTWARE WILL ACCOMPLISH THE INTENDED RESULTS OR THAT THE SOFTWARE OR ITS USE WILL NOT RESULT IN INJURY OR DAMAGE.
# THE USER ASSUMES RESPONSIBILITY FOR ALL LIABILITIES, PENALTIES, FINES, CLAIMS, CAUSES OF ACTION, AND COSTS AND EXPENSES, CAUSED BY, RESULTING FROM OR ARISING OUT OF,
# IN WHOLE OR IN PART THE USE, STORAGE OR DISPOSAL OF THE SOFTWARE.
##############################################################################################################################################################################

from ctypes import c_int, c_double, c_char_p, c_void_p, POINTER, CFUNCTYPE, cdll
import numpy as np
import sys

import TasmanianConfig
import TasmanianSG
TasmanianInputError = TasmanianConfig.TasmanianInputError

pLibCTSG = cdll.LoadLibrary(TasmanianConfig.__path_libcaddons__)

type_lpnmodel = CFUNCTYPE(None, c_int, POINTER(c_double), c_int, POINTER(c_double), c_int, POINTER(c_int))
type_scsmodel = CFUNCTYPE(None, c_int, c_int, POINTER(c_double), c_int, POINTER(c_double), c_int, POINTER(c_int))
type_icsmodel = CFUNCTYPE(None, c_int, c_int, POINTER(c_double), c_int, c_int, POINTER(c_double), c_int, POINTER(c_int))

pLibCTSG.tsgLoadNeededPoints.argtypes = [c_int, type_lpnmodel, c_void_p, c_int, POINTER(c_int)]

pLibCTSG.tsgConstructSurrogateNoIGSurplus.argtypes = [type_scsmodel, c_int, c_int, c_int, c_void_p, c_double, c_char_p,
                                                      c_int, POINTER(c_int), c_char_p, POINTER(c_int)]
pLibCTSG.tsgConstructSurrogateNoIGAniso.argtypes = [type_scsmodel, c_int, c_int, c_int, c_void_p, c_char_p, c_int, POINTER(c_int), c_char_p, POINTER(c_int)]
pLibCTSG.tsgConstructSurrogateNoIGAnisoFixed.argtypes = [type_scsmodel, c_int, c_int, c_int, c_void_p, c_char_p,
                                                         POINTER(c_int), POINTER(c_int), c_char_p, POINTER(c_int)]

pLibCTSG.tsgConstructSurrogateWiIGSurplus.argtypes = [type_icsmodel, c_int, c_int, c_int, c_void_p, c_double, c_char_p,
                                                      c_int, POINTER(c_int), c_char_p, POINTER(c_int)]
pLibCTSG.tsgConstructSurrogateWiIGAniso.argtypes = [type_icsmodel, c_int, c_int, c_int, c_void_p, c_char_p, c_int, POINTER(c_int), c_char_p, POINTER(c_int)]
pLibCTSG.tsgConstructSurrogateWiIGAnisoFixed.argtypes = [type_icsmodel, c_int, c_int, c_int, c_void_p, c_char_p,
                                                         POINTER(c_int), POINTER(c_int), c_char_p, POINTER(c_int)]

pLibCTSG.tsgLoadUnstructuredDataL2.argtypes = [POINTER(c_double), c_int, POINTER(c_double), c_double, c_void_p]

def tsgLnpModelWrapper(oUserModel, iSizeX, pX, iSizeY, pY, iThreadID, pErrInfo):
    '''
    DO NOT CALL DIRECTLY
    This is callback from C++, see TasGrid::loadNeededPoints()

    Creates an interface between a user callable object and
    the Tasmanian callback from C++.
    The callback passes in two raw-arrays with given sizes
    (that correspond to grid inputs and outputs)
    and the id of the running thread.
    The raw-array is wrapped in a numpy array structure and given
    to the user, the result is written to the output array.

    oUserModel: is callable object, e.g., function or lambda

    iSizeX: number of entries in x, equal to getNumDimensions()
    pX:     raw-array corresponding to the model inputs
    iSizeY: number of entries in y, equal to getNumOutputs()
    pY:     raw-array corresponding to the model outputs

    iThreadID: the id of the running thread
    '''
    pErrInfo[0] = 1
    aX = np.ctypeslib.as_array(pX, (iSizeX,))
    aY = np.ctypeslib.as_array(pY, (iSizeY,))
    aResult = oUserModel(aX, iThreadID)
    if aResult.shape != (iSizeY,):
        if TasmanianConfig.enableVerboseErrors:
            print("ERROR: incorrect model output dimensions, should be (iNumOutputs,)")
        return
    aY[0:iSizeY] = aResult[0:iSizeY]
    pErrInfo[0] = 0


def loadNeededPoints(callableModel, grid, iNumThreads = 1):
    '''
    Wrapper to TasGrid::loadNeededPoints(), non-overwrite version.

    If the grid has needed points, the callableModel will be called
    for each grid point (i.e., model input) and the resulting values
    will be loaded in the grid.

    callableModel: is callable object, e.g., function or lambda
        The object must accept two inputs and give one output:
        aY = callableModel(aX, iThreadID)

        aX: is a one dimensional numpy.ndarray with size equal
            to the number of model inputs

        iThreadID: is the ID of the thread executing the model,
            always between 0 and iNumThreads -1
            Two simultaneous calls to callableModel() will always
            have different ids.

        Return: aY must be a one dimensional numpy.ndarray with
            size equal to the number of model outputs

        Note: if iNumThreads > 1, then callableModel() must be thread-safe.

    grid: must be an instance of Tasmanian.SparseGrid()
        model values will be loaded in the grid

    iNumThreads: integer, if greater than 1 the model will be called
        in parallel from multiple threads.
        See TasGrid::loadNeededPoints().

    '''
    iOverwrite = 0 # do not overwrite
    pErrorCode = (c_int * 1)()
    pLibCTSG.tsgLoadNeededPoints(iOverwrite,
                                 type_lpnmodel(lambda nx, x, ny, y, tid, err : tsgLnpModelWrapper(callableModel, nx, x, ny, y, tid, err)),
                                 grid.pGrid, iNumThreads, pErrorCode)
    if pErrorCode[0] != 0:
        raise TasmanianInputError("loadNeededPoints", "An error occurred during the call to Tasmanian.")

def reloadLoadedPoints(callableModel, grid, iNumThreads = 1):
    '''
    Wrapper to TasGrid::loadNeededPoints(), overwrite version.

    Clears any pending refinement (i.e., needed points) and overwrites the model
    values associated with the existing loaded points.

    The inputs are identical to Tasmanian.loadNeededPoints().

    '''
    iOverwrite = 1 # do overwrite
    pErrorCode = (c_int * 1)()
    pLibCTSG.tsgLoadNeededPoints(iOverwrite,
                                 type_lpnmodel(lambda nx, x, ny, y, tid, err : tsgLnpModelWrapper(callableModel, nx, x, ny, y, tid, err)),
                                 grid.pGrid, iNumThreads, pErrorCode)
    if pErrorCode[0] != 0:
        raise TasmanianInputError("reloadLoadedPoints()", "An error occurred during the call to Tasmanian.")

###############################################################################
################### Construct Surrogate #######################################
###############################################################################

def tsgScsModelWrapper(oUserModel, iNumSamples, iNumDims, pX, iNumOuts, pY, iThreadID, pErrInfo):
    '''
    DO NOT CALL DIRECTLY
    This is callback from C++, see TasGrid::constructSurrogate()

    Handles the case of batch models:
    oUserModel: user defined model that takes a two dimensional
        array of inputs and returns a two dimensional array
        of outputs, and a thread ID

    iNumSamples: number of samples in the batch
    iNumDims:    number of model inputs per sample
    iNumOuts:    number of model outputs per sample

    pX and pY are 2D arrays of c_doubles with size iNumSamples
        times iNumDims and iNumOuts respectively

    iThreadID: the id of the running thread

    '''
    pErrInfo[0] = 1
    aX = np.ctypeslib.as_array(pX, (iNumSamples,iNumDims))
    aY = np.ctypeslib.as_array(pY, (iNumSamples,iNumOuts))
    aResult = oUserModel(aX, iThreadID)
    if aResult.shape != (iNumSamples, iNumOuts):
        if TasmanianConfig.enableVerboseErrors:
            print("ERROR: incorrect model output dimensions, should be (iNumSamples, iNumOutputs)")
        return
    aY[0:iNumSamples, 0:iNumOuts] = aResult[0:iNumSamples, 0:iNumOuts]
    pErrInfo[0] = 0

def tsgIcsModelWrapper(oUserModel, iNumSamples, iNumDims, pX, iHasGuess, iNumOuts, pY, iThreadID, pErrInfo):
    '''
    DO NOT CALL DIRECTLY
    This is callback from C++, see TasGrid::constructSurrogate()

    See tsgScsModelWrapper(), the only difference is that
    the user model oUserModel() takes two arrays,
    one with the inputs and one with the initial guess.
    The initial guess could be empty.

    iHasGuess is a boolean that determines whether an
    initial guess has been loaded in pY.
    '''
    pErrInfo[0] = 1
    aX = np.ctypeslib.as_array(pX, (iNumSamples,iNumDims))
    aY = np.ctypeslib.as_array(pY, (iNumSamples,iNumOuts))
    if (iHasGuess == 0): # no guess
        aResult = oUserModel(aX, np.empty([0,0], np.float64), iThreadID)
    else:
        aResult = oUserModel(aX, aY, iThreadID)
    if aResult.shape != (iNumSamples, iNumOuts):
        if TasmanianConfig.enableVerboseErrors:
            print("ERROR: incorrect model output dimensions, should be (iNumSamples, iNumOutputs)")
        return
    aY[0:iNumSamples, 0:iNumOuts] = aResult[0:iNumSamples, 0:iNumOuts]
    pErrInfo[0] = 0

def constructAnisotropicSurrogate(callableModel, iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid,
                                  sDepthType, liAnisotropicWeightsOrOutput,
                                  liLevelLimits = [], bUseInitialGuess = False,
                                  sCheckpointFilename = ""):
    '''
    Construct a surrogate model to the callableModel using
    anisotropic refinement until the iMaxPoints is reached.

    See the documentation for TasGrid::constructSurrogate()
    This is wrapper around the anisotropic refinement variant.

    callableModel is a function (or lambda) that returns a two dimensional
        numpy.ndarray of outputs for the model,
        If bUseInitialGuess is true, the model has to take two inputs,
        otherwise the model uses one.
        The first input is a two dimensional numpy.ndarray of inputs
        similar to TasmanianSparseGrid.evaluateBatch().
        The second input corresponds to the initial guess, the size will
        either match the expected output or will be empty, if no
        guess can be computed.

    iMaxPoints: is a positive integer indicating the maximum number of points
        that the grid will have.
    iMaxParallel: is a positive integer indicating the number of simultaneous
        calls to the user model, i.e., the number of threads.
    iMaxSamplesPerCall: maximum number of samples that will be given to
        a single call to the user model.
    grid: must be an instance of Tasmanian.SparseGrid() with either
        global, sequence or Fourier grid.
    sDepthType: the type used for refinement, see TasGrid::constructSurrogate()

    liAnisotropicWeightsOrOutput: is either an output to use to determine
        the model anisotropy or a list/typle/ndarray of user selected
        anisotropic weights.
    liLevelLimits: same as in all other refinement calls, the refinement will
        never add points below the given level in the diven direction
        even if the budget has not been reached yet.

    sCheckpointFilename: filename to use to checkpoint the algorithm so that
        construction can proceed from a saved point in case of a crash.
    '''
    iNumDims = grid.getNumDimensions()
    pLevelLimits = None
    if (len(liLevelLimits) > 0):
        if (len(liLevelLimits) != iNumDims):
            raise TasmanianInputError("liLevelLimits", "ERROR: invalid number of level limits, must be equal to the grid dimension")
        pLevelLimits = (c_int*iNumDims)()
        for iI in range(iNumDims):
            pLevelLimits[iI] = liLevelLimits[iI]

    if (sys.version_info.major == 3):
        sDepthTypeCtypes = bytes(sDepthType, encoding='utf8')
        if (sCheckpointFilename):
            sCheckpointFilename = bytes(sCheckpointFilename, encoding='utf8')
    else:
        sDepthTypeCtypes = sDepthType

    pCPFname = None
    if (sCheckpointFilename):
        pCPFname = c_char_p(sCheckpointFilename)

    pErrorCode = (c_int * 1)()

    if (((sys.version_info.major == 3) and isinstance(liAnisotropicWeightsOrOutput, int))
            or ((sys.version_info.major == 2) and isinstance(liAnisotropicWeightsOrOutput, (int, long)))):
        # will call the algorithm to dynamically estimate the weights
        iOutput = liAnisotropicWeightsOrOutput

        if (bUseInitialGuess):
            pLibCTSG.tsgConstructSurrogateWiIGAniso(
                type_icsmodel(lambda nx, nd, x, f, ny, y, tid, err : tsgIcsModelWrapper(callableModel, nx, nd, x, f, ny, y, tid, err)),
                iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
                c_char_p(sDepthTypeCtypes), iOutput, pLevelLimits, pCPFname, pErrorCode)
        else:
            pLibCTSG.tsgConstructSurrogateNoIGAniso(
                type_scsmodel(lambda nx, nd, x, ny, y, tid, err : tsgScsModelWrapper(callableModel, nx, nd, x, ny, y, tid, err)),
                iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
                c_char_p(sDepthTypeCtypes), iOutput, pLevelLimits, pCPFname, pErrorCode)
    else:
        # weights are set by the user
        pAnisoWeights = None
        if (len(liAnisotropicWeightsOrOutput) > 0):
            if (sDepthType in TasmanianSG.lsTsgCurvedTypes):
                iNumWeights = 2*grid.getNumDimensions()
            else:
                iNumWeights = grid.getNumDimensions()
            if (len(liAnisotropicWeightsOrOutput) != iNumWeights):
                raise TasmanianInputError("liAnisotropicWeightsOrOutput", "ERROR: wrong number of liAnisotropicWeightsOrOutput, sType '{0:s}' needs {1:1d} weights but len(liAnisotropicWeightsOrOutput) == {2:1d}".format(sDepthType, iNumWeights, len(liAnisotropicWeightsOrOutput)))
            else:
                aAWeights = np.array([liAnisotropicWeightsOrOutput[i] for i in range(iNumWeights)], np.int32)
                pAnisoWeights = np.ctypeslib.as_ctypes(aAWeights)

        if (bUseInitialGuess):
            pLibCTSG.tsgConstructSurrogateWiIGAnisoFixed(
                type_icsmodel(lambda nx, nd, x, f, ny, y, tid, err : tsgIcsModelWrapper(callableModel, nx, nd, x, f, ny, y, tid, err)),
                iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
                c_char_p(sDepthTypeCtypes), pAnisoWeights, pLevelLimits, pCPFname, pErrorCode)
        else:
            pLibCTSG.tsgConstructSurrogateNoIGAnisoFixed(
                type_scsmodel(lambda nx, nd, x, ny, y, tid, err : tsgScsModelWrapper(callableModel, nx, nd, x, ny, y, tid, err)),
                iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
                c_char_p(sDepthTypeCtypes), pAnisoWeights, pLevelLimits, pCPFname, pErrorCode)

    if pErrorCode[0] != 0:
        raise TasmanianInputError("constructSurplusSurrogate", "An error occurred during the call to Tasmanian.")


def constructSurplusSurrogate(callableModel, iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid,
                              fTolerance, sRefinementType, iOutput = -1,
                              liLevelLimits = [], bUseInitialGuess = False,
                              sCheckpointFilename = ""):
    '''
    Construct a surrogate model to the callableModel using surplus refinement
    until either the iMaxPoints or the tolerance are reached.

    See Tasmanian.constructAnisotropicSurrogate() for all matchin inputs,
    except the grid has to be local polynomial and the refinement proceeds
    until the budget is exhausted or the fTolerance is reached.
    The sRefinementType is the same as in the call to local surplus refinement,
    same with the iOutput.
    '''
    iNumDims = grid.getNumDimensions()
    pLevelLimits = None
    if (len(liLevelLimits) > 0):
        if (len(liLevelLimits) != iNumDims):
            raise TasmanianInputError("liLevelLimits", "ERROR: invalid number of level limits, must be equal to the grid dimension")
        pLevelLimits = (c_int*iNumDims)()
        for iI in range(iNumDims):
            pLevelLimits[iI] = liLevelLimits[iI]

    if (sys.version_info.major == 3):
        sRefinementType = bytes(sRefinementType, encoding='utf8')
        if (sCheckpointFilename):
            sCheckpointFilename = bytes(sCheckpointFilename, encoding='utf8')

    pCPFname = None
    if (sCheckpointFilename):
        pCPFname = c_char_p(sCheckpointFilename)

    pErrorCode = (c_int * 1)()

    if (bUseInitialGuess):
        pLibCTSG.tsgConstructSurrogateWiIGSurplus(
            type_icsmodel(lambda nx, nd, x, f, ny, y, tid, err : tsgIcsModelWrapper(callableModel, nx, nd, x, f, ny, y, tid, err)),
            iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
            fTolerance, c_char_p(sRefinementType), iOutput, pLevelLimits, pCPFname, pErrorCode)
    else:
        pLibCTSG.tsgConstructSurrogateNoIGSurplus(
            type_scsmodel(lambda nx, nd, x, ny, y, tid, err : tsgScsModelWrapper(callableModel, nx, nd, x, ny, y, tid, err)),
            iMaxPoints, iMaxParallel, iMaxSamplesPerCall, grid.pGrid,
            fTolerance, c_char_p(sRefinementType), iOutput, pLevelLimits, pCPFname, pErrorCode)

    if pErrorCode[0] != 0:
        raise TasmanianInputError("constructSurplusSurrogate", "An error occurred during the call to Tasmanian.")


def loadUnstructuredDataL2(points, model_data, tolerance, grid):
    '''
    Wrapper around TasGrid::loadUnstructuredDataL2(), see the C++ reference.

    points is 2D ndarray with shape(num_data, grid.getNumDimensions())
    model_data is 2D ndarray with shape(num_data, grid.getNumOutputs())
    grid is an instance of TasmanianSparseGrid
    '''
    if len(points.shape) != 2 or points.shape[1] != grid.getNumDimensions():
        raise TasmanianInputError("points", "ERROR: points must be a 2D numpy.ndarray with points.shape[1] == grid.getNumDimensions()")
    if len(model_data.shape) != 2 or model_data.shape[1] != grid.getNumOutputs():
        raise TasmanianInputError("model_data", "ERROR: model_data must be a 2D numpy.ndarray with model_data.shape[1] == grid.getNumOutputs()")
    if points.shape[0] != model_data.shape[0]:
        raise TasmanianInputError("model_data", "ERROR: mismatch between shape[0] of points and model_data")
    if not hasattr(grid, "TasmanianSparseGridObject"):
        raise TasmanianInputError("grid", "ERROR: grid must be an instance of TasmanianSparseGrid")

    num_data = points.shape[0]
    pLibCTSG.tsgLoadUnstructuredDataL2(np.ctypeslib.as_ctypes(points.reshape((points.size,))), num_data,
                                       np.ctypeslib.as_ctypes(model_data.reshape((model_data.size,))), tolerance, grid.pGrid)



#!/usr/bin/env python

# This is a very fragile script, be VERY careful when you run it
# in short, this script takes ../SparseGrids/TasmanianSparseGrid.hpp/cpp files
# parses them and outputs corresponding .h, windows.h, and windows.hpp files
# thus, only the ../SparseGrids/TasmanianSparseGrid.hpp/cpp are ever edited
# while with one go of the script the windows and C headers are generated
# the output is written to this folder, careful not to overwrite things


with open("../../SparseGrids/TasmanianSparseGrid.hpp") as tsghpp:
    lTsgHpp = tsghpp.readlines()

with open("../../SparseGrids/TasmanianSparseGridWrapC.cpp") as tsgcpp:
    lTsgCpp = tsgcpp.readlines()

lTsgH =[]

iStage = 0
for sLine in lTsgHpp:
    sLine = sLine.replace('\n', '')
    if (iStage == 0): # copyright statement
        lTsgH.append(sLine + "\n")
        if ("*/" in sLine):
            iStage = 1
    elif (iStage == 1): # ifndef statement
        lTsgH.append("\n")
        lTsgH.append("#ifndef __TASMANIAN_SPARSE_GRID_H\n")
        lTsgH.append("#define __TASMANIAN_SPARSE_GRID_H\n")
        lTsgH.append("\n")
        lTsgH.append("// ------------ C Interface for TasmanianSparseGrid -------------- //\n")
        lTsgH.append("// NOTE: you need to explicitly call the constructor and destructor\n")
        lTsgH.append("//       in all cases, void *grid is a pointer to a C++ class\n")
        lTsgH.append("\n")
        iStage = 2

iStage = 0
for sLine in lTsgCpp:
    if (iStage == 0):
        if ('extern "C"' in sLine):
            iStage = 1
    elif (iStage == 1):
        # if this is the definition of a function
        if (sLine.startswith("//")):
            lTsgH.append(sLine)
        elif (("void* tsg" in sLine) or
            ("void *tsg" in sLine) or
            ("void * tsg" in sLine) or
            ("void tsg" in sLine) or
            ("int* tsg" in sLine) or
            ("int *tsg" in sLine) or
            ("int * tsg" in sLine) or
            ("int tsg" in sLine) or
            ("double* tsg" in sLine) or
            ("double *tsg" in sLine) or
            ("double * tsg" in sLine) or
            ("double tsg" in sLine) or
            ("char* tsg" in sLine) or
            ("char *tsg" in sLine) or
            ("char * tsg" in sLine) or
            ("char tsg" in sLine)):
            sDef = sLine.split("{")[0] + ";"

            lTsgH.append(sDef + "\n")
            #print(sDef)

lTsgH.append("\n")
lTsgH.append("#endif\n")

with open("TasmanianSparseGrid.h", 'w') as outp:
    outp.writelines(lTsgH)



import unittest
import TasmanianSG
import sys, os
import numpy as np

class TestTasCommon(unittest.TestCase):
    '''
    The class contains functions and untilities that will be used in many
    places throughout the testing suite. The class needs to run a "dummy"
    test from within the class constructor, otherwise an object cannot
    be initialized from outside of this module.
    '''
    def __init__(self):
        unittest.TestCase.__init__(self, "testNothing")

    def testNothing(self):
        pass

    def compareGrids(self, gridA, gridB, bTestRuleNames = True):
        '''
        Compaes two grids by checking points, weights, and several evaluate.
        The evaluates are done on a canonical [-1, 1] interval.

        The test passes if the grids are mathematically idential.

        bTestRuleNames should be True in most cases, i.e., check if the names
        of the rule in each grid matches. The exception is when comparing the
        GaussPatterson grids build with custom file and sRule = "gauss-patterson"
        '''
        # test basic number of points data
        self.assertEqual(gridA.getNumDimensions(), gridB.getNumDimensions(), "error in getNumDimensions()")
        self.assertEqual(gridA.getNumOutputs(), gridB.getNumOutputs(), "error in getNumOutputs()")
        self.assertEqual(gridA.getNumPoints(), gridB.getNumPoints(), "error in getNumPoints()")
        self.assertEqual(gridA.getNumLoaded(), gridB.getNumLoaded(), "error in getNumLoaded()")
        self.assertEqual(gridA.getNumNeeded(), gridB.getNumNeeded(), "error in getNumNeeded()")
        self.assertEqual(gridA.isUsingConstruction(), gridB.isUsingConstruction(), "error in isUsingConstruction()")

        if (gridA.getNumPoints() == 0): # emptry grid, nothing else to check
            return

        # load test points (canonical domain only), make sure to avoid grid points, e.g., 1/2, 1/4, etc.
        if (gridA.getNumDimensions() == 2):
            mX1 = np.array([1.0/3.0, 1.0/6.0])
            mX2 = np.array([-1.0/3.0, 1.0/6.0])
            mX3 = np.array([-1.0/5.0, -1.0/7.0])
            mX4 = np.array([1.0/7.0, 1.0/5.0])
            mX5 = np.array([-1.0/7.0, -1.0/13.0])
        elif (gridA.getNumDimensions() == 3):
            mX1 = np.array([1.0/3.0, 1.0/6.0, -1.0/5.0])
            mX2 = np.array([-1.0/3.0, 1.0/6.0, 1.0/6.0])
            mX3 = np.array([-1.0/5.0, -1.0/7.0, 1.0/3.0])
            mX4 = np.array([1.0/7.0, 1.0/5.0, 2.0/3.0])
            mX5 = np.array([-1.0/7.0, -1.0/13.0, -2.0/3.0])

        aBatchPoints = np.row_stack([mX1, mX2, mX3, mX4, mX5])

        pA = gridA.getPoints()
        pB = gridB.getPoints()
        np.testing.assert_equal(pA, pB, "Points not equal", True)

        pA = gridA.getLoadedPoints()
        pB = gridB.getLoadedPoints()
        np.testing.assert_equal(pA, pB, "Loaded not equal", True)

        pA = gridA.getNeededPoints()
        pB = gridB.getNeededPoints()
        np.testing.assert_equal(pA, pB, "Needed not equal", True)

        # test rule data
        self.assertEqual(gridA.getAlpha(), gridB.getAlpha(), "error in getAlpha()")
        self.assertEqual(gridA.getBeta(), gridB.getBeta(), "error in getBeta()")
        self.assertEqual(gridA.getOrder(), gridB.getOrder(), "error in getOrder()")

        if (bTestRuleNames):
            self.assertEqual(gridA.getRule(), gridB.getRule(), "error in getRule()")
            self.assertEqual(gridA.getCustomRuleDescription(), gridB.getCustomRuleDescription(), "error in getCustomRuleDescription()")

        self.assertEqual(gridA.isGlobal(), gridB.isGlobal(), "error in isGlobal()")
        self.assertEqual(gridA.isSequence(), gridB.isSequence(), "error in isSequence()")
        self.assertEqual(gridA.isLocalPolynomial(), gridB.isLocalPolynomial(), "error in isLocalPolynomial()")
        self.assertEqual(gridA.isWavelet(), gridB.isWavelet(), "error in isWavelet()")
        self.assertEqual(gridA.isFourier(), gridB.isFourier(), "error in isFourier()")

        # weights
        pA = gridA.getQuadratureWeights()
        pB = gridB.getQuadratureWeights()
        np.testing.assert_equal(pA, pB, "Quadrature not equal", True)

        pA = gridA.getInterpolationWeights(mX1)
        pB = gridB.getInterpolationWeights(mX1)
        np.testing.assert_equal(pA, pB, "Interpolation test 1 not equal", True)

        pA = gridA.getInterpolationWeights(mX2)
        pB = gridB.getInterpolationWeights(mX2)
        np.testing.assert_equal(pA, pB, "Interpolation test 2 not equal", True)

        pA = gridA.getInterpolationWeights(mX3)
        pB = gridB.getInterpolationWeights(mX3)
        np.testing.assert_equal(pA, pB, "Interpolation test 3 not equal", True)

        # evaluate (values have been loaded)
        if (gridA.getNumLoaded() > 0):
            pA = gridA.evaluate(mX4)
            pB = gridB.evaluate(mX4)
            np.testing.assert_almost_equal(pA, pB, 12, "Interpolation test 4 not equal", True)

            pA = gridA.evaluate(mX5)
            pB = gridB.evaluate(mX5)
            np.testing.assert_almost_equal(pA, pB, 12, "Interpolation test 5 not equal", True)

            pA = gridA.integrate()
            pB = gridB.integrate()
            np.testing.assert_almost_equal(pA, pB, 12, "Integration test not equal", True)

            pA = gridA.evaluateBatch(aBatchPoints)
            pB = gridB.evaluateBatch(aBatchPoints)
            np.testing.assert_almost_equal(pA, pB, 12, "Interpolation test 6 (batch) not equal", True)

            pA = gridA.getHierarchicalCoefficients()
            pB = gridB.getHierarchicalCoefficients()
            np.testing.assert_almost_equal(pA, pB, 12, "getHierarchicalCoefficients() not equal", True)

        # domain transforms
        self.assertEqual(gridA.isSetDomainTransfrom(), gridB.isSetDomainTransfrom(), "error in isSetDomainTransfrom()")

        pA = gridA.getDomainTransform()
        pB = gridB.getDomainTransform()
        np.testing.assert_equal(pA, pB, "Domain test no equal", True)

        self.assertEqual(gridA.isSetConformalTransformASIN(), gridB.isSetConformalTransformASIN(), "error in isSetConformalTransformASIN()")

        pA = gridA.getLevelLimits()
        pB = gridB.getLevelLimits()
        np.testing.assert_equal(pA, pB, "Level limit test no equal", True)

        pA = gridA.getConformalTransformASIN()
        pB = gridB.getConformalTransformASIN()
        np.testing.assert_equal(pA, pB, "Conformal transform ASIN not equal", True)

    def loadExpN2(self, grid):
        '''
        If there are needed points, load the points with exp(-sum(x)**2) where
        x is each of the needed points.

        The function is needed to test whether read/write correctly works on the
        loaded values, so some values have to be easily loaded even if they
        are not meaningful as in the convergence/correctness tests.
        '''
        if (grid.getNumNeeded() == 0): return
        mPoints = grid.getNeededPoints()
        iOut = grid.getNumOutputs()
        iDim = grid.getNumDimensions()
        grid.loadNeededPoints(np.column_stack([np.exp(-np.sum(mPoints**2, axis=1)) for i in range(iOut)]))

    def checkPoints(self, aPoints, lMustHave, lMustNotHave):
        '''
        aPoints is 1D array of points
        lMustHave and lMustNotHave are the points to test
        Example: checkPoints(aPoints[:,0], [0.0, 1.0, -1.0], [0.5, -0.5])
        '''
        for x in lMustHave:
            self.assertTrue((np.sum(np.where(np.abs(aPoints - x) < 0.001, 1, 0)) > 0), 'did not properly limit level, did not find {0:1.5e}'.format(x))
        for x in lMustNotHave:
            self.assertTrue((np.sum(np.where(np.abs(aPoints - x) < 0.001, 1, 0)) == 0), 'did not properly limit level, did find {0:1.5e}'.format(x))
